# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from megatron.bridge.models.gemma.gemma2_bridge import Gemma2Bridge  # noqa: F401
from megatron.bridge.models.gemma.gemma2_provider import (
    Gemma2ModelProvider,
    Gemma2ModelProvider2B,
    Gemma2ModelProvider9B,
    Gemma2ModelProvider27B,
)
from megatron.bridge.models.gemma.gemma3_bridge import Gemma3ModelBridge  # noqa: F401
from megatron.bridge.models.gemma.gemma3_provider import (
    Gemma3ModelProvider,
    Gemma3ModelProvider1B,
    Gemma3ModelProvider4B,
    Gemma3ModelProvider12B,
    Gemma3ModelProvider27B,
)
from megatron.bridge.models.gemma.gemma_bridge import GemmaBridge  # noqa: F401
from megatron.bridge.models.gemma.gemma_provider import (
    CodeGemmaModelProvider2B,
    CodeGemmaModelProvider7B,
    GemmaModelProvider,
    GemmaModelProvider2B,
    GemmaModelProvider7B,
)


__all__ = [
    "GemmaModelProvider",
    "GemmaModelProvider2B",
    "GemmaModelProvider7B",
    "CodeGemmaModelProvider2B",
    "CodeGemmaModelProvider7B",
    "Gemma2ModelProvider",
    "Gemma2ModelProvider2B",
    "Gemma2ModelProvider9B",
    "Gemma2ModelProvider27B",
    "Gemma3ModelProvider",
    "Gemma3ModelProvider1B",
    "Gemma3ModelProvider4B",
    "Gemma3ModelProvider12B",
    "Gemma3ModelProvider27B",
]
