# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Megatron Bridge Recipe Configurations

This module exposes all recipe configurations from all model families.
"""

from megatron.bridge.recipes.deepseek import *
from megatron.bridge.recipes.gemma import *
from megatron.bridge.recipes.gemma3_vl import *
from megatron.bridge.recipes.glm import *
from megatron.bridge.recipes.glm_vl import *
from megatron.bridge.recipes.gpt import *
from megatron.bridge.recipes.gpt_oss import *
from megatron.bridge.recipes.llama import *
from megatron.bridge.recipes.ministral3 import *
from megatron.bridge.recipes.moonlight import *
from megatron.bridge.recipes.nemotronh import *
from megatron.bridge.recipes.olmoe import *
from megatron.bridge.recipes.qwen import *
from megatron.bridge.recipes.qwen_vl import *
