# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from megatron.bridge.models.nemotronh.nemotron_h_bridge import NemotronHBridge
from megatron.bridge.models.nemotronh.nemotron_h_provider import (
    Nemotron3NanoProvider,
    NemotronHModel4BProvider,
    NemotronHModel8BProvider,
    NemotronHModel47BProvider,
    NemotronHModel56BProvider,
    NemotronHModelProvider,
    NemotronHModelProvider4B,
    NemotronHModelProvider8B,
    NemotronHModelProvider47B,
    NemotronHModelProvider56B,
    NemotronNano9Bv2Provider,
    NemotronNano12Bv2Provider,
    NemotronNanoModelProvider9Bv2,
    NemotronNanoModelProvider12Bv2,
)


__all__ = [
    "NemotronHBridge",
    "NemotronHModelProvider",
    "NemotronHModelProvider4B",
    "NemotronHModelProvider8B",
    "NemotronHModelProvider47B",
    "NemotronHModelProvider56B",
    "NemotronNanoModelProvider9Bv2",
    "NemotronNanoModelProvider12Bv2",
    "NemotronHModel4BProvider",
    "NemotronHModel8BProvider",
    "NemotronHModel47BProvider",
    "NemotronHModel56BProvider",
    "NemotronNano9Bv2Provider",
    "NemotronNano12Bv2Provider",
    "Nemotron3NanoProvider",
]
