# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Gemma2 models
from .gemma2 import (
    gemma2_2b_peft_config,
    gemma2_2b_pretrain_config,
    gemma2_2b_sft_config,
    gemma2_9b_peft_config,
    gemma2_9b_pretrain_config,
    gemma2_9b_sft_config,
    gemma2_27b_peft_config,
    gemma2_27b_pretrain_config,
    gemma2_27b_sft_config,
)

# Gemma3 models
from .gemma3 import (
    gemma3_1b_peft_config,
    gemma3_1b_pretrain_config,
    gemma3_1b_sft_config,
)


__all__ = [
    # Gemma2 models
    "gemma2_2b_pretrain_config",
    "gemma2_9b_pretrain_config",
    "gemma2_27b_pretrain_config",
    "gemma2_2b_sft_config",
    "gemma2_9b_sft_config",
    "gemma2_27b_sft_config",
    "gemma2_2b_peft_config",
    "gemma2_9b_peft_config",
    "gemma2_27b_peft_config",
    # Gemma3 models
    "gemma3_1b_pretrain_config",
    "gemma3_1b_sft_config",
    "gemma3_1b_peft_config",
]
